/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * dataset - functions for managing data sets. Usually one instance
 * represents all the data loaded from a single data file.
 */

#include <ctype.h>
#include <math.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <gtk/gtkmain.h>

#include "dataset.xh"
#include "main.xh"
#include "util.xh"
#include "stringutil.xh"

static void free_dataset(my_dataset_t *dataset);

static my_time_t parse_timestamp(const char **p) {
   return (my_time_t)(parse_double_num(p) * TIME_MULTIPLIER);
}

static void parse_line(my_dataset_t *dataset, char *line, my_time_t timeoff, int is_millis) {
   const char *p = line;

   my_time_t timestamp = parse_timestamp(&p);
   if(!isspace(*p) && *p) {
      chomp(line);
      fprintf(stderr, "Bad data '%s' in line '%s'\n", p, line);
      return;
   }

   if(is_millis) {
      timestamp /= 1000;
   }

   const int lineidx = dataset->time_count;

   my_time_arr_ensure(dataset->time, lineidx + 1);

   dataset->time[lineidx] = timestamp + timeoff;

   double prev[100]; //##TODO## not fixed

   for(int i=0; *(p = skipws(p)); ++i) {
      my_data_t num;
      do {
	 num = parse_double_num(&p);
	 if(!isspace(*p) && *p) {
	    if(*p != ':') {
	       // it's not a time field.. complain
	       chomp(line);
	       fprintf(stderr, "Bad data '%s' in line '%s'\n", p, line);
	    }
	    p = skipws(skipnonws(p));
	    continue;
	 }
	 break;
      } while(1);

      if(!*p) break;

      my_dataset_item_arr_ensure(dataset->item, i+1);
      //if(!dataset->item[i].data) printf("new %d at %Ld\n", i, timestamp);
      my_data_arr_ensure(dataset->item[i].data, lineidx + 1);

      if(num >= 84.99 && num <= 85.01) { num = NAN; } else
      {
	 // check minmax
	 if(dataset->item[i].max < num) dataset->item[i].max = num;
	 if(dataset->item[i].min > num) dataset->item[i].min = num;

	 // check diff minmax
	 if(lineidx > 0) {
	    double diff = (num - prev[i]) * TIME_MULTIPLIER / (double)(dataset->time[lineidx] - dataset->time[lineidx - 1]);
	    if(diff > 0) {
	       if(dataset->item[i].min_diff < diff) dataset->item[i].min_diff = diff;
	    } else {
	       if(dataset->item[i].max_diff < diff) dataset->item[i].max_diff = diff;
	    }
	 }
      }

      dataset->item[i].data[lineidx] = num;
      prev[i] = num;
   }
}

#define MAX_LINE_LEN 500

my_dataset_t *read_dataset_int(const char *file, int maxlines, progress_cb_t *cb, gpointer user_data) {
   my_dataset_t *dataset = zmalloc_c(sizeof(my_dataset_t));

   dataset->item_dflt.min = INFINITY;
   dataset->item_dflt.max = -INFINITY;
   dataset->item_dflt.data_dflt = NAN;

   FILE *f = fopen(file, "rt");
   if(f == NULL) return NULL;

   file = dataset->path = g_strdup(file);

   const char *filename = strrchr(file, '/');
   if(filename) {
      // slash found, jump to next letter
      filename++;
   } else {
      // no slash, use full name
      filename = file;
   }

   dataset->name = filename;

   time_t pd_l1 = time(0);

   char line[MAX_LINE_LEN];
   char *labels[100]; //##TODO## non-fixed
   int numlabels = 0;

   // fseek(f, -55 * 10000L, SEEK_END); fgets(line, (int)sizeof(line), f); // skip first partial line

   my_time_t ts;

   do {
      if(maxlines >= 0) {
	 if(maxlines-- == 0) {
	    fclose(f);

	    for(int i=0; i<numlabels; ++i) {
	       free(labels[i]);
	    }
	    free_dataset(dataset);
	    return NULL;
	 }
      }
      if(!fgets(line, (int)sizeof(line), f)) {
	 fclose(f);

	 for(int i=0; i<numlabels; ++i) {
	    free(labels[i]);
	 }
	 free_dataset(dataset);
	 return NULL;
      }

      if(numlabels == 0) {
	 // try to detect column labels
	 const char *x = skipws(line);
	 if(*x == '#') {
	    x++;
	    x = skipws(x);
	 }

	 // this "detection" is too simple but works in basic cases
	 if(isalpha(*x)) {
	    // labels found -> detect separator - check for tabs, multiple spaces or default to single space
	    const char *sep;
	    if(strchr(x, '\t')) {
	       sep = "\t";
	    } else if(strstr(x, "  ")) {
	       sep = "  ";
	    } else {
	       sep = " ";
	    }

	    char *next;
	    while((next = strstr(x, sep))) {
	       labels[numlabels++] = g_strndup(x, (size_t)(next - x));
	       x = skipws(next+1);
	    }
	    labels[numlabels++] = g_strdup(x);

	    continue;
	 }
      }

      strip_comments(line);

      const char *p = line;
      ts = parse_timestamp(&p);
      if(p == line || (!isspace(*p) && *p)) {
	 chomp(line);
	 fprintf(stderr, "Ignoring bad line '%s'\n", line);
	 continue;
      }

      break;
   } while(1);

   my_time_t timeoff = 0;

   dataset->timetype = TIMETYPE_EPOCH;
   dataset->ts = ts;
   int is_millis = 0;

   if(ts > 100000000000LL * TIME_MULTIPLIER) {
      is_millis = 1;

   } else if(ts < 100000000LL * TIME_MULTIPLIER) {
      dataset->timetype = TIMETYPE_START;
   }

   // get last modification time and assume that's also the time of the last entry
   struct stat st;
   if(stat(file, &st)) {
      perror("read_dataset(): failed to stat file");
   }

   dataset->last_modified = st.st_mtime * TIME_MULTIPLIER;

   // try to get the relative time of the last entry
   if(dataset->timetype == TIMETYPE_START) {
      long origoff = ftell(f);

      fseek(f, -MAX_LINE_LEN - 1L, SEEK_END);
      char eline[MAX_LINE_LEN];
      my_time_t te = 0;
      fgets(eline, (int)sizeof(eline), f); // skip first partial line
      while(fgets_strip_comments(eline, (int)sizeof(eline), f)) {
	 const char *p = eline;
	 my_time_t tmp = parse_timestamp(&p);
	 if(p > eline) {
	    // update only if successful
	    te = tmp;
	 }
      }

      timeoff = st.st_mtime * TIME_MULTIPLIER - te;

      fseek(f, origoff, SEEK_SET);
   }

   if(maxlines >= 0) {
      if(maxlines == 0) {
	 maxlines = 1;
      }

      for(int i=0; i<maxlines; ++i) {
	 parse_line(dataset, line, timeoff, is_millis);

	 if(!fgets_strip_comments(line, (int)sizeof(line), f)) break;
      }

      int origcount = dataset->time_count;

      // scan end of file also in case there are more data columns there

      fseek(f, -MAX_LINE_LEN - 1L, SEEK_END);
      fgets(line, (int)sizeof(line), f); // skip first partial line
      while(fgets_strip_comments(line, (int)sizeof(line), f)) {
	 parse_line(dataset, line, timeoff, is_millis);
      }

      dataset->te = dataset->time[dataset->time_count-1];

      // restore time count
      dataset->time_count = origcount;

   } else {
      int idx = 0;

      do {
	 parse_line(dataset, line, timeoff, is_millis);

	 if(!(idx % 5000)) {
	    if(!(idx % 20000)) {
	       if(cb) {
		  cb(((double)ftell(f)) / st.st_size, user_data);
	       }
	    }
	    while(gtk_events_pending()) {
	       gtk_main_iteration();
	    }

	    if(quit) {
	       if(gtk_main_level()) gtk_main_quit();
	       dataset->item_count = 0;
	       break;
	    }

	 }
	 idx++;

      } while(fgets_strip_comments(line, (int)sizeof(line), f));

      dataset->te = dataset->time[dataset->time_count-1];
   }

   fclose(f);

   // make sure all items have same amount of elements and insert labels
   for(int i=0; i<dataset->item_count; ++i) {
      my_data_arr_ensure(dataset->item[i].data, dataset->time_count);
      if(i < numlabels-1) {
	 dataset->item[i].name = labels[i+1];
      }
   }

   if(dataset->item_count == 0) {
      free_dataset(dataset);
      return NULL;
   }

   time_t pd_l2 = time(0);
   printf("  loading time: %ld seconds\n", pd_l2 - pd_l1);

   ref_dataset(dataset);

   return dataset;
}

static void free_dataset(my_dataset_t *dataset) {
   for(int i=0; i<dataset->item_count; ++i) {
      my_data_arr_free(dataset->item[i].data);
      if(dataset->item[i].name) {
	 free(dataset->item[i].name);
      }
   }

   my_dataset_item_arr_free(dataset->item);

   if(dataset->path) {
      free(dataset->path);
   }

   bzero(dataset, sizeof(my_dataset_t));
   free(dataset);
}

void ref_dataset(my_dataset_t *dataset) {
   ++dataset->refcount;
}

int unref_dataset(my_dataset_t *dataset) {
   if(--dataset->refcount <= 0) {
      printf("Freeing dataset %s\n", dataset->name);
      free(dataset);
      return 1;
   }
   return 0;
}
